/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*                                 sangers                               */
/*                                                                       */
/*************************************************************************/

/**************************************************************************
  FILE           : $Source: /home/patricio/pry/src/sangers.c,v $
  SHORTNAME      : sangers.c

  PURPOSE        : Extracción de las componentes principales mediante 
                   red de Sanger
  NOTES          : 

  AUTHOR         : Patricio García Báez (pgarcia AT ull.es)
                   Departamento de Estadística, I.O. y Computación
                   Universidad de La Laguna
                   38271 La Laguna - S/C de Tenerife
                   Canarias - España
  DATE           : 21/04/1994
  LICENCE        : View file COPYING

  CHANGED BY     : 
  RCS VERSION    : $Revision: 0.7 $
  LAST CHANGE    : $Date: 2010/12/20 13:00:00 $

    Copyright (c) 2006 Patricio García Báez y la Universidad de La Laguna

**************************************************************************/

/**************************************************************************
  DESCRIPTION    : Implementa red neuronal de Sanger (o GHA) para 
                   la extración de las componentes principales.
		     Ver [Hertz, Krog, Palmer, 1991].

                   Para ello recibe parámetros tanto desde la línea de 
                   comandos (ver -h) como de un fichero de ejecuciones 
                   (.rus).

                   Emite por la salida de error estandar diversa
                   información sobre la evolución del aprendizaje. 

  DEPENDENCIES   : Hace uso de hermat (libreria de funciones para 
                   la lectura y escritura de matrices de floats).

  COMMENTS       : 

  BUGS           : 

**************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "hermat.h"
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <math.h>


/*******************************/
/* FUNCIONES Y MACROS DE APOYO */
/*******************************/


#ifndef max
#define max(x,y)        ((x)>(y)?(x):(y))
#endif

#ifndef min
#define min(x,y)        ((x)<(y)?(x):(y))
#endif


elem modulo(unsigned n, vector v)
    /* Calcula el módulo de un vector */
{
    unsigned i;
    elem *ptev, sum;

    for (i = n, ptev = v, sum = 0.0; i--; ptev++)
	if (!isnan(*ptev))
	    sum += (*ptev) * (*ptev);
    sum = sqrt(sum);

    return (sum);
}


elem media(unsigned n, vector v)
    /* Calcula la media de un vector */
{
    unsigned i, miss;
    elem *ptev, med;

    for (i = n, ptev = v, med = 0.0, miss = 0; i--; ptev++)
	if (isnan(*ptev))
	    miss++;
	else
	    med += *ptev;
    if (n > miss)
	med /= (n - miss);

    return (med);
}


vector med0(unsigned n, vector v)
    /* Ajusta la media de un vector a 0 */
{
    unsigned i;
    elem *ptev, med;

    med = media(n, v);

    for (i = n, ptev = v; i--; ptev++)
	if (!isnan(*ptev))
	    *ptev -= med;

    return (v);
}


vector norm(unsigned n, vector v)
    /* Normaliza un vector a modulo = 1 (no usado) */
{
    unsigned i;
    elem *ptev, sum;

    for (i = n, ptev = v, sum = 0; i--; ptev++)
	if (!isnan(*ptev))
	    sum += (*ptev) * (*ptev);
    sum = sqrt(sum);

    for (i = n, ptev = v; i--; ptev++)
	if (!isnan(*ptev))
	    *ptev /= sum;

    return (v);
}


/*******************************/
/* AJUSTA LA MEDIA Y NORMALIZA */
/*******************************/

matriz ajusta(unsigned m, unsigned n, matriz mat)
{
    unsigned  i;
    vector *ptvm;

    for (i = m, ptvm = mat; i--; ptvm++) {
	*ptvm = med0(n, *ptvm);
	/* *ptvm = norm(n, *ptvm); */
#ifdef DEBUG
	printf("ajust[%u]=%f\n", m - i, **ptvm);
#endif
    }
    return (mat);
}


/************************************/
/* CALCULO DE LA MEDIA DE UN VECTOR */
/************************************/

void mediav(unsigned m, unsigned n, matriz mat, vector v)
    /* Calcula la media de un grupo de vectores */
{
    unsigned i, j, *miss, *pmiss;
    vector indi, indv;

    miss = pmiss = (unsigned *) calloc(n, sizeof(unsigned));

    for (j = m; j--;)
	for (indi = mat[j], indv = v, i = n, pmiss = miss; i--; indv++, indi++, pmiss++)
	    if (isnan(*indi))
		(*pmiss)++;
	    else
		*indv += *indi;

    for (i = n, indv = v, pmiss = miss; i--;)
	(*indv++) /= (m - (*pmiss++));

    free(miss);
}


/***************************************/
/* COLOCA A 0 LA MEDIA DE CADA COLUMNA */
/***************************************/

matriz mean0(unsigned m, unsigned n, matriz mat, vector vmed)
{
    unsigned  j, i;
    vector *ptvm;
    elem *ptem, *ptev;

    for (j = m, ptvm = mat; j--;)
	for (i = n, ptem = (*ptvm++), ptev = vmed; i--; ptem++, ptev++)
	    if (!isnan(*ptem))
		*ptem -= *ptev;

    return (mat);
}


/**********************************/
/* INICIALIZACIÓN DE PESOS        */
/**********************************/

matriz inic_sanger(unsigned m, unsigned n, matriz w, unsigned seed)
    /* Asignación de pesos aleatorios */
{
    unsigned i, j;
    vector *ptvw;
    elem *ptew;

    srand(seed);
#ifdef DEBUG2
    fprintf(stderr, "Comp: rand=%d\tRAND_MAX=%d\t", rand(), RAND_MAX);
#endif

    for (i = m, ptvw = w; i--; ptvw++)
	for (j = n, ptew = *ptvw; j--;) {
	    float     r = (unsigned) rand() & RAND_MAX;
	    r /= RAND_MAX;            /* 0.0 .. 1.0 */
	    *ptew++ = r;
#ifdef DEBUG2
	    fprintf(stderr, "Comp: r=%f\t", r);
#endif
	}
    return (w);
}


/*****************************/
/* SALIDA DE SANGER          */
/*****************************/

vector sal_sanger(unsigned n, unsigned n_a, vector mat, matriz w, vector v)
    /* Cálculo de la salida de sanger */
{
    unsigned i, j;
    vector *ptvw;
    elem *ptev, *ptew, *ptem;

#ifdef DEBUG2
    fprintf(stderr, "\nEntra: %d\t%d\n", n, n_a);
#endif

    for (i = n_a, ptev = v; i--;)
	(*ptev++) = 0;

    for (j = n, ptvw = w, ptem = mat; j--; ptem++, ptvw++)
	if (!isnan(*ptem))
	    for (i = n_a, ptev = v, ptew = *ptvw; i--;) {
		(*ptev++) += (*ptew++) * (*ptem);
#ifdef DEBUG2
		fprintf(stderr, "\nComp: *ptew=%f\t*ptem=%f\t*ptev=%f\n", *(ptew-1) , (*ptem), *(ptev-1));
#endif
	    }

    /* #ifdef DEBUG2
       fprintf(stderr, "\nIte.: -1 \n");
       for (i = n_a, ptev = v; i--;)
       fprintf(stderr, "%f\t", *ptev++);
       #endif */
    return (v);
}


/**********************************/
/* RECONSTRUCCIÓN DE LAS MUESTRAS */
/**********************************/

vector recsanger(unsigned n, unsigned n_a, matriz w, vector v, vector rec)
    /* Cálculo de la reconstrucción */
{
    unsigned i, j;
    vector *ptvw;
    elem *ptev, *ptew, *pter;

    for (j = n, ptvw = w, pter = rec; j--; ptvw++) {
	register elem sum = 0.0;
	for (i = n_a, ptev = v, ptew = *ptvw; i--;) 
	    sum += (*ptew++) * (*ptev++);
	(*pter++) = sum;
    }
    return (rec);
}


elem mse(vector v, vector rec, unsigned n)
    /* Cálculo de la media de errores cuadráticos */
{
    unsigned i;
    elem *ptev, *pter, sum2 = 0.0;

    for (i = n, ptev = v, pter = rec; i--; ptev++, pter++)
	if (!isnan(*ptev))
	    sum2 += ((*ptev) - (*pter)) * ((*ptev) - (*pter));

    return (sum2);
}

elem mse_rec(unsigned m, unsigned n, unsigned n_a, matriz mat, matriz w, vector v, vector rec)
    /* Cálculo error de reconstrucción */
{
    unsigned i;
    vector *ptvm;
    elem err_rec = 0.0;
    
    for (i = m, ptvm = mat; i--; ptvm++) {
	v = sal_sanger(n, n_a, *ptvm, w, v);
	rec = recsanger(n, n_a, w, v, rec);
	err_rec += mse(*ptvm, rec, n);
    }
    err_rec /= m;
    return (err_rec);
}


/*****************************/
/* APRENDIZAJE DE SANGER     */
/*****************************/

void display_apren_sanger(unsigned ciclo, float nu, float err_act, unsigned errrec, float err_rec)
    /* Visualización de mensajes de evolución del aprendizaje */
{
    fprintf(stderr, "Número de ciclo:      %d\n", ciclo);
    fprintf(stderr, "Ratio Actual:         %g\n", nu);
    fprintf(stderr, "ModIncWAct:           %g\n", err_act * nu * nu);
    fprintf(stderr, "ModIncWAct/RatioAct:  %g\n", err_act);
    if (errrec)
	fprintf(stderr, "MSERec (%u CP):        %g\n", errrec, err_rec);
    fprintf(stderr, "---------------------------------------------\n\n");
}


matriz entren_sanger(unsigned m, unsigned n, unsigned n_a, matriz mat, matriz w, float nuin, unsigned nmi, float error, unsigned T2, unsigned errrec)
    /* Entrenamiento de sanger ppd */
{
    unsigned l, h, j, i, k;
    vector *ptvm, *ptvw;
    elem *ptem, *ptew, *ptew2, *ptev, *ptev2;
    float nu = nuin, err_act;
    vector v = reserva_vector(n_a);
    vector rec = reserva_vector(n);
#ifdef DEBUG0
    vector med_w = reserva_vector(n_a);
    elem *ptemw;
#endif

    /* Mientras no se cumplan las condiciones de parada */
    for (l = nmi, err_act = error; (l--) && (err_act >= error); nu -= nuin / nmi) {

#ifdef DEBUG0
	for (i=n_a, ptemw = med_w; i--;) {
	    (*ptemw++) = 0;
	}
#endif

	/* Para todos los patrones */
	for (h = m, ptvm = mat, err_act = 0.0; h--; ptvm++) {
	    v = sal_sanger(n, n_a, *ptvm, w, v);
#ifdef DEBUG2
	    fprintf(stderr, "\nIte.: %u\tMuest.: %u\n", nmi - l, m - h);
	    for (i = n_a, ptev = v; i--;)
		fprintf(stderr, "%f\t", *ptev++);
#endif
	    /* Para todos los pesos */
	    for (j = n, ptvw = w, ptem = *ptvm; j--; ptvw++, ptem++)
		if (!isnan(*ptem)) {
#ifdef DEBUG0
		    ptemw = med_w;
#endif
		    for (i = n_a, ptew = *ptvw, ptev = v; i--;) {
			elem inc_w = *ptem;
			for (k = n_a - i, ptev2 = v, ptew2 = *ptvw; k--;)
			    inc_w -= (*ptev2++) * (*ptew2++);
			inc_w *= (*ptev++);
			(*ptew++) += nu * inc_w;
			err_act += inc_w * inc_w;
#ifdef DEBUG0
			(*ptemw++)+= inc_w * inc_w;
#endif
		    }
		}
	}
	err_act /= n * m * n_a;


	/* Cálculo e impresión de logs */
#ifdef DEBUG0
	for (i = n_a, ptemw = med_w; i--; ptemw++) {
	    *ptemw /= n * m;
	}

	fprintf(stderr, "ItAct : %u  RatioAct: %g  ModIncWAct/RatioAct: %g\nModIncwAct/RatioAct: ", nmi-l, nu, err_act);
	for (i = n_a, ptemw = med_w; i--;)
	    fprintf(stderr, "%g ", (*ptemw++));
	fprintf(stderr, "\n");	
#endif

	if (T2 && ((nmi - l - 1) % T2 == 0)) {
	    elem err_rec = 0.0;
	    if (errrec)
		err_rec = mse_rec(m, n, errrec, mat, w, v, rec);
	    display_apren_sanger(nmi-l, nu, err_act, errrec, err_rec);
	}
    }

    libera_vector(v);
    libera_vector(rec);
#ifdef DEBUG0
    libera_vector(med_w);
#endif
    return (w);
}


/*****************************/
/* APRENDIZAJE DE SANGER     */
/*****************************/

matriz apren_sanger(unsigned m, unsigned n, unsigned n_a, char *fichin1, char *fichin2, char *fichout1, unsigned aj, unsigned covar, vector *v_med, matriz w, unsigned seed, float nu, unsigned nmi, float max_err_a, unsigned T2, unsigned errrec)
{
    matriz mat;

    FILE *fi1, *fi2, *fo;

    mat = reserva_matriz(m, n);
    if ((fi1 = fopen(fichin1, "r")) == NULL)
	Err("No se pudo abrir fichero de patrones de entrenamiento");
    lee_matriz(fi1, m, n, mat);
    fclose(fi1);

    if ((strcmp(fichin2, "") == 0) || (strcmp(fichin2, "RAMDOM") == 0)) {
	/* pesos aleatorios */
	w = inic_sanger(n, n_a, w, seed);
    } else {
	if ((fi2 = fopen(fichin2, "r")) == NULL)
	    Err("No se pudo abrir fichero de pesos");
	lee_matriz(fi2, n, n_a, w);
	fclose(fi2);
    }

    if (aj)
	mat = ajusta(m, n, mat);

    if (covar) {
	*v_med = reserva_vector(n);
	mediav(m, n, mat, *v_med);
	mat = mean0(m, n, mat, *v_med);
    }

    errrec = min(n_a, errrec);

    w = entren_sanger(m, n, n_a, mat, w, nu, nmi, max_err_a, T2, errrec);


    if ((fo = fopen(fichout1, "w")) == NULL)
	Err("No se pudo abrir fichero de pesos");
    escribe_matriz(fo, n, n_a, w);
    
    fclose(fo);
    libera_matriz(m, mat);

    return (w);
}


/*****************************/
/* EJECUCIÓN DE SANGER       */
/*****************************/

#define varianza(sum, sum2, n) (((sum2) / (n)) - (((sum) * (sum)) / ((n) * (n))))


void display_ejec_sanger(unsigned m, unsigned errvar, unsigned errrec, elem sumx, elem sumx2, elem sumy, elem sumy2, elem sumer2)
    /* Visualización de mensajes del resultado de la ejecución */
{
    if (errvar) {
	float varin = varianza(sumx, sumx2, m);
	float varout = varianza(sumy, sumy2, m);

	fprintf(stderr,"Varianza en la entrada:                       %g\n", varin);
	fprintf(stderr,"Varianza en la/s %u primera/s CP/s:            %g\n", errvar, varout);
	fprintf(stderr,"%% var. representada por la/s %u CP/s:          %f%%\n", errvar, (varout * 100) / varin);
    }

    if (errrec) {
	sumer2 /= m;
	fprintf(stderr,"MSE reconstrucción de la/s %u primera/s CP/s:  %g\n", errrec, sumer2);
    fprintf(stderr, "---------------------------------------------\n");
    }
}


void ejec_sanger(unsigned m, unsigned n, unsigned n_a, char* fichin1, char* fichout2, unsigned covar, unsigned aj, matriz w, vector v_med, unsigned errvar, unsigned errrec)
    /* Cambia a una base reducida w los vectores de la matriz leida */
    /* del disco y el resultado lo coloca a disco tambiën */
{
    unsigned i, leidas;
    vector mat = reserva_vector(n);
    vector des = reserva_vector(n_a);
    vector rec = NULL;
    float mod, sumx = 0.0, sumx2 = 0.0, sumy = 0.0, sumy2 = 0.0, sumer2 = 0.0;

    FILE *fi1, *fo;


    if ((fi1 = fopen(fichin1, "r")) == NULL)
	Err("No se pudo abrir fichero de patrones de test");

    if ((fo = fopen(fichout2, "w")) == NULL)
	Err("No se pudo abrir fichero de patrones de salida");

    errvar = min(n_a, errvar);
    errrec = min(n_a, errrec);
    if (errrec)
	rec = reserva_vector(n);

    leidas = lee_matriz(fi1, 1, n, &mat);
    for (i = m; (i--) && (leidas == 1);) {
	/* Ajusta entrada */
	if (aj)
	    ajusta(1,n, &mat);

	if (covar)
	    mean0(1, n, &mat, v_med);
	
	des = sal_sanger(n, n_a, mat, w, des);
#ifdef DEBUG
	fprintf(stderr,"leidas =%u, pr = %f\t", leidas,*mat);
#endif
	
	if (errvar) {
	    sumx += media(n, mat);
	    mod = modulo(n, mat);
	    sumx2 += mod * mod;
	    sumy += media(errvar, des);
	    mod = modulo(errvar, des);
	    sumy2 += mod * mod;
	}
	
	if (errrec) {
	    rec = recsanger(n, errrec, w, des, rec);
	    sumer2 += mse(mat, rec, n);
	}
	
	escribe_matriz(fo, 1, n_a, &des);

	if (i)
	    leidas = lee_matriz(fi1, 1, n, &mat);
    }

    if (leidas != 1)
	m -= i + 1;

    display_ejec_sanger(m, errvar, errrec, sumx, sumx2, sumy, sumy2, sumer2);

    libera_vector(mat);
    libera_vector(des);

    fclose(fi1);
    fclose(fo);
}


/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

char *progname;


void usage(unsigned nmi, float max_err_a, float nu,  unsigned errvar, unsigned errrec, unsigned seed)
{
    fprintf(stderr, "Usar: %s [-mv] [-a rat_inic] [-e int_log] [-i nmi] [-r cp_mse] [-s cp_var] [-t val_par] fich_rus\n", progname);
    fprintf(stderr, "Extracción de las componentes principales mediante red de Sanger.\n");
    fprintf(stderr, "Opciones:\n");
    fprintf(stderr, "  -V                    Imprime versión del programa.\n");
    fprintf(stderr, "  -m                    No utilizar entradas con media por filas a 0.\n");
    fprintf(stderr, "  -v                    Utilizar entradas con media por columnas a 0.\n");
    fprintf(stderr, "  -a rat_inic           Ratio inicial de aprendizaje de la red de Sanger (%g),\n\t\t\t desciende linealmente a 0 en nmi.\n", nu);
    fprintf(stderr, "  -c cp_var             Número de componentes para cálculo de la varianza\n\t\t\t acumulada (%u).\n", errvar);
    fprintf(stderr, "  -e int_log            Intervalo entre el que se generan logs (nmi - 1).\n");
    fprintf(stderr, "  -i nmi                Número máximo de iteraciones (nmi) en la red (%u).\n", nmi);
    fprintf(stderr, "  -r cp_mse             Número de componentes para cálculo del promedio\n\t\t\t del error cuadratico (mse) de reconstrucción (%u).\n", errrec);
    fprintf(stderr, "  -s seed               Valor de la semilla para números aleatorios (%u).\n", seed);
    fprintf(stderr, "  -t val_par            Valor de parada sobre sobre el promedio del cuadrado\n\t\t\t de los incrementos de pesos entre el ratio actual\n\t\t\t en un ciclo (%g).\n", max_err_a);

    exit(EX_USAGE);
}


#define FL 100
#define TODASCP 999999

int  main(int argc, char *argv[])
{
    extern char *optarg;
    extern int optind;
    int opt;

    unsigned n;             /* número de muestras por vector (N)               */
    unsigned m;             /* número de vectores de aprendizaje (M)           */
    unsigned m_r;           /* numero de vectores a reducir (si no = M)
		             * (M_R)  */
    unsigned n_a;           /* número de autovectores a calcular (N_A)         */
    unsigned nmi = 200;     /* número máximo de iteraciones  para cada
			      * cálculo */
    unsigned covar = 0;     /* indica si se quiere aprendizaje con media
			      * por columnas a 0 o no */
    unsigned aj = 1;             /* indica si se quiere la media de cada patron a 0 */
    unsigned errvar = TODASCP; /* indica núm. de comp. para calc. % variaza acumulada */
	                     /* 0 = no calcular, TODASCP = todas las CP */
    unsigned errrec = TODASCP; /* indica núm. de comp. para calc. err. reconstrucción */
	                     /* 0 = no calcular, TODASCP = todas las CP */
    unsigned seed = 1;      /* Semilla de números aleatorios */
    unsigned T2;            /* Intervalo entre el que se generan informes de error */
    unsigned inT2 = 0;      /* El usuario ha entrado T2 */
    unsigned r, n_runs;

    float  nu = 0.3;         /* ratio del algoritmo de sanger                   */
    float max_err_a = 0;     /* maximo error admitido para sanger               */

    char fichin1[FL];        /* fichero con las muestras (MxN)                  */
    char fichin2[FL] = "";   /* fichero de pesos de entrada (N_AxN)             */
    char fichout1[FL];       /* fichero de pesos de salida (N_AxN)              */
    char fichout2[FL];       /* fichero de muestras reducidas de salida
			      * (M_RxN_A) */
    char sterror[FL] = "";     /* String de error */

    FILE     *frun = NULL;
    matriz    w;
    vector    v_med = NULL;

    /* Leer argumentos desde la línea de comandos */
    progname = argv[0];

    while ((opt = getopt(argc, argv, "Vvmi:t:a:c:e:r:s:h")) != EOF)
	switch (opt) {
	case 'V':
	    printf("%s: %s version %s\n", progname, PACKAGE_NAME, PACKAGE_VERSION);
	    exit(EX_OK);
	    break;
	case 'v':
	    covar = 1;
	    break;
	case 'm':
	    aj = 0;
	    break;
	case 'i':
	    sscanf(optarg, "%u", &nmi);
	    break;
	case 't':
	    sscanf(optarg, "%f", &max_err_a);
	    break;
	case 'a':
	    sscanf(optarg, "%f", &nu);
	    break;
	case 'c':
	    sscanf(optarg, "%u", &errvar);
	    break;
	case 'e':
	    sscanf(optarg, "%d", &T2);
	    inT2 = 1;
	    break;
	case 'r':
	    sscanf(optarg, "%u", &errrec);
	    break;
	case 's':
	    sscanf(optarg, "%u", &seed);
	    break;
	case 'h':
	case '?':
	    usage(nmi, max_err_a, nu, errvar, errrec, seed);
	    break;
	}

    if (argc == (optind + 1)) {
	/* abrir el fichero de ejecuciones para lectura */
	if ((frun = fopen(argv[optind++], "r")) == NULL) {
	    perror(strcat(strcat(strcat(sterror, progname), ": "),argv[optind - 1]));
	    exit(EX_NOINPUT);
	}
    } else
	usage(nmi, max_err_a, nu, errvar, errrec, seed);


    if (!inT2)
	T2 = nmi - 1;


    fscanf(frun, "%d", &n_runs);  /* busca el número de ejecuciones */

    /*******************************************/
    /* Comienzo del bucle de trabajo Principal */
    /*******************************************/

    for (r = 0; r < n_runs; r++) {
	/* Lee información sobre las ejecuciones */
	fscanf(frun, "%u %u %s %u %s %s %u %s",
	       &n,              /* número de entradas (N) */
	       &m,              /* número de patrones de aprendizaje (M) */
	       fichin1,         /* nombre del fichero de patrones (aprendizaje+test) (M_R*N)*/
	       &n_a,            /* número de neuronas de salida (características) (C)*/
	       fichout1,        /* nombre del fichero de pesos de salida (N*C)*/
	       fichout2,        /* nombre del fichero de resultados (M_R*C) */
	       &m_r,            /* número de patrones a testear (M_R) */
	       fichin2);        /* nombre del fichero de pesos de entrada (N*C)*/


	w = reserva_matriz(n, n_a);

	w = apren_sanger(m, n, n_a, fichin1, fichin2, fichout1, aj, covar, &v_med, w, seed, nu, nmi, max_err_a, T2, errrec);

	ejec_sanger(m_r, n, n_a, fichin1, fichout2, covar, aj, w, v_med, errvar, errrec);

	if (v_med)
	    libera_vector(v_med);
	libera_matriz(n, w);
    }

    return (EX_OK);
}
